% Function to evaluate the code using 1-NN as ground truth
% Used to evaluate LSH, Anti sparse, qoLSH code
% 
% Usage: [result] = eval_coding (a_x, a_xq, x, xq, F, gnd, cfg, name)
% where
%   a_x        input vector code
%   a_xq       query vector code
%   x          input vector
%   xq         query vector
%   F          Projection frame used by LSH
%   gnd        ground truth: 1-NN
%   cfg        configuration object
%   name       identifier - used for display/logging 
%
% This software is governed by the CeCILL license under French law and
% abiding by the rules of distribution of free software. 
% See http://www.cecill.info/licences.en.html
%
% This package was written by Raghavendran Balu, Teddy Furon, Herve Jegou
% Copyright (C) INRIA
% Last change: February 2014. 

function [result] = eval_coding (a_x, a_xq, x, xq, F, gnd, cfg, name)

L = size(a_x,1);
d = size(x,1);
n = size(x,2);
nq = size(xq,2);

if L < 10
  code_x = compact_bits (a_x, 'uint64');
  code_xq = compact_bits (a_xq, 'uint64');
  fprintf ('Unique codes %d -> %d.\n  Database: %d\n  Query:    %d\n', ...
    d, L, length (unique(code_x)), length (unique(code_xq)));
%  break;
%  error ('End')
end

% Compute Hamming distance

code_x = compact_bits (a_x, 'uint8');
code_xq = compact_bits (a_xq, 'uint8');

tic
%[~, idx_ham] = yael_hamming_kmin ( code_x, code_xq, cfg.k); function not included, memory efficient
[~, idx_ham] = yael_kmin ( single (yael_hamming (code_x, code_xq)) , cfg.k);
fprintf ('Hamming distance compute time: %.3f\n',toc);

rank_ham = compute_ranks (gnd, idx_ham);


% Explicit reconstruction for product+sign
rec_x =  yael_fvecs_normalize (F * a_x);
rec_xq =  yael_fvecs_normalize (F * a_xq);

recpinv_x =  yael_fvecs_normalize (pinv(F)' * a_x);
recpinv_xq =  yael_fvecs_normalize (pinv(F)' * a_xq);

% Compute reconstruction
[idx_rec, ~] = yael_nn (rec_x, rec_xq, cfg.k);
rank_rec = compute_ranks (gnd, idx_rec);

[idx_recpinv, ~] = yael_nn (recpinv_x, recpinv_xq, cfg.k);
rank_recpinv = compute_ranks (gnd, idx_recpinv);

% Asymetric performance
[idx_rec_asym, ~] = yael_nn (rec_x, xq, cfg.k);
rank_rec_asym = compute_ranks (gnd, idx_rec_asym);

% Asymetric (filtered)  performance
idx_rec_asym_filt = zeros(size(idx_rec_asym));

for qi = 1:nq
[idx_rec_asym_filt(:, qi),~] = yael_nn (rec_x(:,idx_ham(:,qi)), xq(:,qi), cfg.k);
idx_rec_asym_filt(:, qi) = idx_ham(idx_rec_asym_filt(:, qi),qi);
end

rank_rec_asym_filt = compute_ranks (gnd, idx_rec_asym_filt);


%% Produce recall@r values
recallk = zeros(length(cfg.rlist), 4);
ri = 0;
for r = cfg.rlist
  ri = ri + 1;
  recallk (ri, 1) = length (find (rank_ham <= r)) / nq;
  recallk (ri, 2) = length (find (rank_rec <= r)) / nq;
  recallk (ri, 3) = length (find (rank_rec_asym <= r)) / nq;  
  recallk (ri, 4) = length (find (rank_rec_asym_filt <= r)) / nq;   
end

%% Compute reconstruction error
mse = zeros(1,2); 
mse(1) = mean (sum(rec_x - x).^2);
mse(2) = mean (sum(recpinv_x - x).^2);

ent = zeros(1,2);
ent(1) = entropy(code_x);
ent(2) = entropy(code_xq);

if cfg.verbose>0
ri=0;
for r = cfg.rlist
  ri = ri + 1; 
  fprintf ('recall@%-5d = %.4f / %.4f / %.4f / %.4f\n', r, recallk(ri, 1), recallk(ri, 2), recallk(ri, 3), recallk(ri, 4));       
end

fprintf ('MSE ps      = %.3f\n', mse(1));
fprintf ('MSE ps/pinv = %.3f\n', mse(2));
fprintf ('Entropy = %.3f\n',ent(1));
end

result(:).name = name;
result(:).recallk = recallk;
result(:).mse = mse;
result(:).ent = ent;

end
