% Generate a set of normalized vectors (frame) using different methods
% This function is used in qolsh for lsh projection matrix generation
% 
% Usage: [A,ind] = gen_frame (d, m, method)
% where
%   d        number of dimensions in the dataset
%   m        projection vector size, decides the code size in LSH 
%   method   
%   'km'     use a kmeans clustering to produce m vectors, then normalize
%   'randn'  just random vectors on the sphere
%   'frame'  the method of Jean-Jacques Fuchs to produce a tight frame
%   'utf'    the method of Teddy to produce a uniform tight frame
%
% This software is governed by the CeCILL license under French law and
% abiding by the rules of distribution of free software. 
% See http://www.cecill.info/licences.en.html
%
% This package was written by Raghavendran Balu, Teddy Furon, Herve Jegou
% Copyright (C) INRIA
% Last change: February 2014.

function [A,ind] = gen_frame (d, m, method)

if strcmp (method, 'km')
  % Data dependent projection matrix
  A = yael_kmeans (single(randn(d, 10000)), m);
  A = yael_fvecs_normalize (A);
  
elseif strcmp (method, 'randn')
  % Random projection matrix 
  A = single (randn (d, m));
  A = yael_fvecs_normalize (A);
  
elseif strcmp (method, 'tightframe')
  % Tight frame
  A = randn (m, d);
  [Q, R] = qr (A);
  A = Q (1:d, :);
elseif strcmp (method, 'framenorm')
  % Tight frame
  A = randn (m, d);
  [Q, R] = qr (A);
  A = Q (1:d, :);
  A = yael_fvecs_normalize (single(A));
elseif strcmp (method, 'utf')
  % Uniform Tight Frame
  if mod(d,2)==1
    error('d must be even');
  end
  if d>m
    error('d must be lower than m')
  end
  A = zeros(d,m);
  indl = randperm(floor(m/2-1));
  ind = indl(1:(d/2));
  indc = 0:(m-1);
  for kl=1:(d/2)
    A(kl,:)=cos(2*pi*indl(kl)*indc/m);
    A(d/2+kl,:)=sin(2*pi*indl(kl)*indc/m);
  end
  A = A * sqrt(2/m);
else
  disp 'Unknown frame construction method'
  assert (0);
end

A = double (A);

end
