% The function improves the LSH code based on reconstruction error at each 
% step by choosing best possible bit coordinate to flip, for maxflips steps.
%
% Usage: [A, Nflips, flc] = qolsh(F, X, maxflips, verbose)
% where
%   F        projection frame
%   X        real vectors dataset
%   maxflips stopping criterion for the coordinate descent
%   verbose  verbosity level
%
% This software is governed by the CeCILL license under French law and
% abiding by the rules of distribution of free software. 
% See http://www.cecill.info/licences.en.html
%
% This package was written by Raghavendran Balu, Teddy Furon, Herve Jegou
% Copyright (C) INRIA
% Last change: February 2014. 

function [A, Nflips, flc] = qolsh(F, X, maxflips, verbose)

flc=0;

d = size (F, 1);
L = size (F, 2);
n = size (X, 2);

assert (size(X, 1) == d);
if nargin < 3, maxflips = L; end
if nargin < 4, verbose = false; end

Ft = F';
gramF = Ft * F;         % pre-computed gram matrix between frame vectors

normsqrF = sum (F.^2)';

A = zeros (L, n);
Nflips = zeros (1, n);

flc = 0;
A = sign(Ft*X);
RX = F*A;
RX_norm = sqrt(sum(RX.^2));
FRX = Ft*RX;
FX = Ft*X;
obj = sum(RX.*X)./RX_norm;
nflips =0;

while 1
    
    newcos = bsxfun(@minus,obj.*RX_norm , 2 * A .* FX);              % compute cos for all bit flips            
    newcos= newcos./sqrt(bsxfun(@minus, bsxfun(@plus, 4 * normsqrF ,RX_norm.^2), ...
    4 * A .* FRX));
          
    [bestflipcos, bestflipidx] = max(newcos);                       % select the best flip   
    cg_idx = bestflipcos>obj;

    if sum(cg_idx)>0
            
        idx = sub2ind(size(A),bestflipidx(cg_idx), find(cg_idx));  % get the index to flip
        A(idx) = -A(idx);
         
        % update the residuals and new cosine 
        RX(:,cg_idx) = F * A(:,cg_idx);
        RX_norm = sqrt(sum(RX.^2));
        FRX(:,cg_idx) = Ft * RX(:,cg_idx);
        obj(cg_idx) = newcos(idx);

        if verbose
            fprintf('%d:%d:%f:%f ',bestflipidx,sum(pflips(bestflipidx,:)), ...
            sum(newcos(bestflipidx,pflips(bestflipidx,:)) - obj(pflips(bestflipidx,:))),...
            mean (sum(yael_fvecs_normalize(RX) - X).^2));
        end
        nflips = nflips + 1;
    else
        if verbose
            fprintf ('# Stop -> reached a local minimum after %d flips\n', nflips);
        end
        break;
    end
    
    if nflips == maxflips                                         % stop at maxflips
        break;
    end
end
if verbose
    fprintf('\n');
end
end



